import numpy as np
import SimpleITK as itk
import pydicom


def dcmread(dicom_path, read_header=False):
    """ Read 2D image data from the DICOM file.

    Args:
        dicom_path (str): path of the dicom file.
        read_header (bool): whether to return the dicom header together
            with the image array.

    Return:
        (ndarray): dicom image array.
        (pydicom.dataset.FileDataset): an instance of FileDataset
        that represents a parsed DICOM file.
    """
    ds = pydicom.dcmread(dicom_path)
    try:
        img = ds.pixel_array
    except Exception:
        img_itk = itk.ReadImage(dicom_path)
        img = itk.GetArrayFromImage(img_itk)
        img = np.squeeze(img)

    if read_header:
        return img, ds
    else:
        return img


def dcminfo(dicom_path):
    """ Read metadata (dicom tags) from DICOM file.

    Refer to pydicom.dcmread

    Args:
        dicom_path (str): path of the dicom file.

    Return:
        (pydicom.dataset.FileDataset): an instance of FileDataset
        that represents a parsed DICOM file.
    """
    ds = pydicom.dcmread(dicom_path)
    return ds



from .hook.hook import Hook
from .hook.logger_hook import LoggerHook
from .hook.cls_logger_hook import ClsLoggerHook
from .runner import Runner


__all__ = [
    'Hook',
    'LoggerHook',
    'ClsLoggerHook',
    'Runner'
]



import cv2


def rgb2gray(src):
    """ Convert a RGB image to grayscale image.

    Args:
        src (ndarray): the input image.

    Returns:
        (ndarray): the converted grayscale image.
    """
    return cv2.cvtColor(src, cv2.COLOR_RGB2GRAY)


def gray2rgb(src):
    """ Convert a grayscale image to RGB image.

    Args:
        src (ndarray): the input image.

    Returns:
        (ndarray): the converted RGB image.
    """
    return cv2.cvtColor(src, cv2.COLOR_GRAY2RGB)



import numpy as np
from .load_annotation import load_bdc_dr_bbox
import medvision as mv


def bdc2dsmd_det_2d(annot_dir,
                    image_dir=None,
                    class2label=None,
                    ignore_label_name=True,
                    replace_ext=lambda x: x):
    # N.B. annotation file name and image file name should be the same
    num_classes = len(class2label) if class2label is not None else 1

    filenames = mv.listdir(annot_dir)
    empty_bboxes = np.zeros((0, 4), dtype=np.float32)
    dsmd = {replace_ext(filename): [empty_bboxes] * num_classes
            for filename in filenames}
    for filename in filenames:
        annot_filepath = mv.joinpath(annot_dir, filename)
        bboxes = load_bdc_dr_bbox(
            annot_filepath,
            lambda x: 0 if ignore_label_name else class2label
        )
        for label, bbox in bboxes:
            bbox = np.array(bbox, dtype=np.float32).reshape(-1, 4)
            if dsmd[replace_ext(filename)][label].shape[0] == 0:
                dsmd[replace_ext(filename)][label] = bbox
            else:
                dsmd[replace_ext(filename)][label] = np.append(
                    dsmd[replace_ext(filename)][label], bbox, axis=0)

    return mv.make_dsmd(dsmd)


if __name__ == '__main__':
    # TODO: move to unittest
    annot_dir = '/mnt/sdb1/tb/internal/label'
    dsmd_path = '/mnt/sdb1/tb/internal/1.csv'

    def replace_ext(x):
        return x.replace('.txt', '.dcm')

    dsmd = bdc2dsmd_det_2d(annot_dir, replace_ext=replace_ext)
    mv.save_dsmd(dsmd_path, dsmd, {'tb': 0}, mode='det')

    dsmd = mv.load_dsmd(dsmd_path, {'tb': 0}, mode='det')
    bboxes = dsmd['002565.dcm'][0]
    image_path = '/mnt/sdb1/tb/internal/tb/002565.dcm'
    image = mv.dcmread_dr(image_path)
    image = mv.normalize_grayscale(image) * 255.0
    image = image.astype(np.uint8)
    mv.imshow_bboxes(image, bboxes, thickness=4)



from .build_object import build_object_from_dict
from .config_util import load_config_cmdline, load_config
from .fileutil import (isdir, isfile, listdir,
                       joinpath, basename, abspath, splitext, parentdir,
                       cp, rm, cptree, rmtree, move, symlink,
                       mkdirs, empty_dir,
                       non_overwrite_cp, copyfiles,
                       GlobMode, glob,
                       find_duplicated_files)
from .init_util import (init_logging, init_random_seed, init_cuda_devices,
                        init_experiment, init_system)
from .multiprocessutil import tqdm_imap
from .priority import Priority, get_priority
from .timer import Timer
from .typeutil import isarrayinstance


__all__ = [
    'build_object_from_dict',

    'load_config_cmdline', 'load_config',

    'isdir', 'isfile', 'listdir',
    'joinpath', 'basename', 'abspath', 'splitext', 'parentdir',
    'cp', 'rm', 'cptree', 'rmtree', 'move', 'symlink',
    'mkdirs', 'empty_dir',
    'non_overwrite_cp', 'copyfiles',
    'GlobMode', 'glob',
    'find_duplicated_files',

    'init_logging', 'init_random_seed', 'init_cuda_devices',
    'init_experiment', 'init_system',

    'tqdm_imap',

    'Priority', 'get_priority',

    'Timer',

    'isarrayinstance',
]



import argparse


def load_config_cmdline(load_default_config,
                        update_config,
                        description='',
                        log_cfg=False):
    """ Load configurations from command line.

    Args:
        load_default_config (function): function to load default
            configurations.
        update_config (function): function to update secondary
            configurations according to primary configurations.
        description(str): refer to argparse.ArgumentParser description.
        log_cfg(bool): whether or not to log configurations.

    Return:
        (yacs CfgNode): loaded configurations

    Example:
        <<< python train.py --config_file config.yaml
    """
    # TODO: add unit test
    # 1. define argument parser
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "--config_file", default=None, type=str,
        help="path to yaml config file"
    )
    parser.add_argument(
        "--opts", default=None, nargs=argparse.REMAINDER,
        help="override config options using command-line",
    )

    args = parser.parse_args()

    # 2 load default configuration
    cfg = load_default_config()

    # 3 load configuration from command line
    if args.config_file is not None:
        cfg.merge_from_file(args.config_file)

    if args.opts is not None:
        cfg.merge_from_list(args.opts)

    # 4. update configurations if primary configuration changes
    update_config(cfg)

    cfg.freeze()

    # 5. log configuration for visual check
    if log_cfg:
        print('Running with config:\n{}'.format(cfg))

    return cfg


def load_config(load_default_config,
                update_config,
                config_file_path=None):
    """ Load configuration.

    Args:
        load_default_config (function): function to load default
            configurations.
        update_config (function): function to update secondary
            configurations according to primary configurations.
        config_file_path (str): configuration file path.

    Return:
        (yacs CfgNode): loaded configurations
    """
    # TODO: add unit test
    cfg = load_default_config()

    if config_file_path is not None:
        cfg.merge_from_file(config_file_path)
    update_config(cfg)

    cfg.freeze()
    return cfg



import time
import numpy as np
import visdom


class VisdomVisualizer:
    """ A wrapper class for visdom.

    Note:
        Original visdom APIs are supported through self.vis.function.
    """
    def __init__(self, env='default', **kwargs):
        self.vis = visdom.Visdom(env=env, **kwargs)
        self.log_text = ''

    def plot(self, name, x, y, **kwargs):
        update = 'append' if self.vis.win_exists(name) else None
        self.vis.line(
            Y=np.array([y]), X=np.array([x]),
            win=name,
            opts=dict(title=name),
            update=update,
            **kwargs
        )

    def show_images(self, name, imgs_, **kwargs):
        self.vis.images(
            imgs_.cpu().numpy(),
            win=name,
            opts=dict(title=name),
            **kwargs
        )

    def log(self, info, win='log'):
        self.log_text += (
            '[{time}] {info} <br>'.format(
                time=time.strftime('%y-%m-%d %H:%M:%S'),
                info=info
            )
        )
        self.vis.text(self.log_text, win)

    def __getattr__(self, name):
        return getattr(self.vis, name)



import logging
import torch
from .logger_hook import LoggerHook
import medvision as mv


class ClsLoggerHook(LoggerHook):
    def __init__(self, interval):
        super().__init__(interval)

    def log(self, runner):
        # 1. logging
        # basic information
        if runner.mode == mv.ModeKey.TRAIN:
            lr_str = ', '.join(
                ['{:.5f}'.format(lr) for lr in runner.current_lr()])
            log_str = 'Epoch [{}][{}/{}]\tlr: {}, '.format(
                runner.epoch + 1, runner.inner_iter + 1,
                len(runner.data_loader), lr_str)
        else:
            log_str = 'Epoch({}) [{}][{}]\t'.format(
                runner.mode, runner.epoch, runner.inner_iter + 1)

        # CUDA memory information
        if torch.cuda.is_available():
            mem = torch.cuda.max_memory_allocated()
            mem_mb = int(mem / (1024 * 1024))
            mem_str = 'memory: {}GB, '.format(mem_mb)
            log_str += mem_str

        # customized intermediate results
        log_items = []
        for name, val in runner.average_meter.output.items():
            if isinstance(val, float):
                val = '{:.4f}'.format(val)
            log_items.append('{}: {}'.format(name, val))
        log_str += ', '.join(log_items)

        logging.info(log_str)

        # 2. TensorboardX
        if runner.mode == mv.ModeKey.TRAIN:
            runner.visualizer.plot(
                'train_loss',
                runner.iter,
                runner.average_meter.output['loss']
            )



from .classification import load_cls_dsmd, save_cls_dsmd
from .detection import load_det_dsmd, save_det_dsmd


def load_dsmd(dsmd_path, class2label=None, mode='cls'):
    """ Load dataset metadata.

    Dataset Metadata is a key-value pairs describing a dataset.
    For example, a dataset metadata can be a dictionary looks like
    {
        'data/1.png': 1,
        'data/2.png': 0,
        ...
    }

    A dataset metadata file defines the mapping from files
    to annotations. Several types of data metadata files are listed below.

    +-------------------------------------------------------+
    | A. Single Label Classification Dataset Metadata File  |
    +-------------------------------------------------------+
    |data/1.png,1                                           |
    |data/2.png,0                                           |
    |...                                                    |
    +-------------------------------------------------------+

    +-------------------------------------------------------+
    | B. Multi-Label Classification Dataset Metadata File   |
    +-------------------------------------------------------+
    |data/1.png,1,0,1                                       |
    |data/2.png,0,1,0                                       |
    |...                                                    |
    +-------------------------------------------------------+

    +-------------------------------------------------------+
    | C. Segmentation Dataset Metadata File                 |
    +-------------------------------------------------------+
    |data/1.png,data/1_mask.png                             |
    |data/2.png,data/2_mask.png                             |
    |...                                                    |
    +-------------------------------------------------------+

    +-------------------------------------------------------+
    | D. Detection Dataset Metadata File A (bbox only)      |
    +-------------------------------------------------------+
    |data/1.png,170,146,397,681,cat                         |
    |data/1.png,473,209,723,673,cat                         |
    |data/1.png,552,167,745,272,dog                         |
    |data/2.png,578,267,624,304,dog                         |
    |data/3.png,,,,,                                        |
    |...                                                    |
    +-------------------------------------------------------+

    +-------------------------------------------------------+
    | E. Detection Dataset Metadata File B (bbox with score)|
    +-------------------------------------------------------+
    |data/1.png,170,146,397,681,0.5,cat                     |
    |data/1.png,473,209,723,673,0.2,cat                     |
    |data/1.png,552,167,745,272,0.7,dog                     |
    |data/2.png,578,267,624,304,0.9,dog                     |
    |data/3.png,,,,,                                        |
    |...                                                    |
    +-------------------------------------------------------+

    Args:
        dsmd_path (str): dataset metadata file path.
        class2label (str or dict): class-to-label file or dict.
        mode (str): dataset mission, can be one of 'cls', 'seg', 'det'.

    Return:
        (OrderedDict): dataset metadata.
    """
    if mode in ['cls', 'seg']:
        return load_cls_dsmd(dsmd_path)
    elif mode == 'det':
        return load_det_dsmd(dsmd_path, class2label)
    else:
        raise ValueError('only support cls, seg, det modes')


def save_dsmd(dsmd_path, data, class2label=None, auto_mkdirs=True, mode='cls'):
    """ Save dataset metadata to specified file.

    Args:
        dsmd_path (str): file path to save dataset metadata.
        data (dict): dataset metadata, refer to 'load_dsmd'.
        class2label (str or dict): class-to-label file or class2label dict.
        auto_mkdirs (bool): If the parent folder of `file_path` does
            not exist, whether to create it automatically.
        mode (str): dataset mission, can be one of 'cls', 'seg', 'det'.
    """
    if mode in ['cls', 'seg']:
        return save_cls_dsmd(dsmd_path, data, auto_mkdirs)
    elif mode == 'det':
        return save_det_dsmd(dsmd_path, data, class2label, auto_mkdirs)
    else:
        raise ValueError('only support cls, seg, det modes')


def load_c2l(c2l_path):
    """ Load class-to-label mapping.

    A class-to-label file defines the mapping from class_names to
    labels, which looks like (Note that the label value starts from 0)

    +------------------------------------------------------+
    | Class-to-Label File                                  |
    +------------------------------------------------------+
    |cat, 0                                                |
    |dog, 1                                                |
    |...                                                   |
    +------------------------------------------------------+

    Args:
        c2l_path (str): class-to-label file.

    Return:
        (OrderedDict): class-to-label mapping.
    """
    return load_dsmd(c2l_path)



import time


class Timer:
    """ Timer class.

    Example:
    >>> import time
    >>> import medvision as mv
    >>> with mv.Timer(description='it takes {:.2f} seconds.'):
    >>>     time.sleep(1)
    it takes 1.00 seconds.
    """
    def __init__(self, description=None):
        self.description = description if description else '{:.2f}'

    def __enter__(self):
        self.t_start = time.time()
        return self

    def __exit__(self, type, value, traceback):
        self.t_end = time.time()
        print(self.description.format(self.t_end - self.t_start))


if __name__ == '__main__':
    with Timer(description='function takes {:.4f} seconds.'):
        time.sleep(1)



import logging.config
import os
import random
import numpy as np
import torch
import yaml
import medvision as mv


def init_logging(log_dir=None, config_file=None):
    if log_dir is None:
        log_dir = os.getcwd()

    if config_file is None:
        config_file = mv.joinpath(mv.parentdir(mv.parentdir(__file__)),
                                  'configs/default_log_config.yaml')

    with open(config_file, 'rt') as f:
        config = yaml.safe_load(f.read())
        config['handlers']['info_file_handler']['filename'] = \
            mv.joinpath(log_dir, 'info.log')
        config['handlers']['error_file_handler']['filename'] = \
            mv.joinpath(log_dir, 'error.log')
        mv.mkdirs(log_dir)
        logging.config.dictConfig(config)


def init_random_seed(seed=0):
    random.seed(seed)
    np.random.seed(seed)

    # code to handle reproducibility issue
    # https://pytorch.org/docs/stable/notes/randomness.html
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def init_cuda_devices(cuda_devices=None):
    if cuda_devices is not None:
        cuda_devices = ','.join([str(d) for d in cuda_devices])
        os.environ['CUDA_VISIBLE_DEVICES'] = cuda_devices


def init_experiment(cfg):
    init_logging(cfg.LOGGING.LOG_DIR)
    init_random_seed(cfg.SYSTEM.RANDOM_SEED)
    init_cuda_devices(cfg.SYSTEM.CUDA_DEVICES)


init_system = init_experiment



from torch.utils.data.dataset import Dataset
import medvision as mv


class ClsDataset(Dataset):
    # TODO: avoid using cfg as argument
    def __init__(self, cfg, mode, build_transform, image_loader):
        self.is_train = (mode == mv.ModeKey.TRAIN)

        self.mode2dsmd = {
            mv.ModeKey.TRAIN: cfg.DATA.TRAIN_DSMD,
            mv.ModeKey.VAL: cfg.DATA.VAL_DSMD,
            mv.ModeKey.TEST: cfg.DATA.TEST_DSMD,
        }
        dsmd_path = self.mode2dsmd[mode]
        assert mv.isfile(dsmd_path) or mv.isdir(dsmd_path)

        if mv.isfile(dsmd_path):  # for a dsmd file
            self.dsmd = mv.load_dsmd(dsmd_path)
            self.filenames = list(self.dsmd.keys())
            self.filepaths = [
                mv.joinpath(cfg.DATA.IMAGE_DIR, filename)
                for filename in self.filenames
            ]
        else:  # for a directory containing test images
            self.dsmd = None
            self.filepaths = mv.listdir(dsmd_path)

        self.transform = build_transform(cfg, self.is_train)
        self.image_loader = image_loader

    def __getitem__(self, index):
        image = self.image_loader(self.filepaths[index])

        if self.transform is not None:
            image = self.transform(image)

        if self.dsmd is None:
            return image
        else:
            return image, self.dsmd[self.filenames[index]]

    def __len__(self):
        return len(self.filepaths)



from .dr_io import DrReadMode, dcmread_dr
from .fileutil import isdicom, isdicomdir
from .io import dcmread, dcminfo


__all__ = [
    'DrReadMode', 'dcmread_dr',
    'isdicom', 'isdicomdir',
    'dcmread', 'dcminfo'
]



import numpy as np
import medvision as mv


def normalize_grayscale(src, to_float=True):
    """ Rescale image intensity.

    Rescale an grayscale image's intensity range to [0.0, 1.0].

    Args:
        src (ndarray): image to be intensity rescaled.
        to_float (bool): whether to convert the input image to float
            before intensity rescale.

    Return:
        (ndarray): intensity rescaled image.
    """
    assert src.ndim == 2
    epsilon = np.finfo(np.float32).eps

    if to_float:
        src = src.astype(np.float32)

    min_val, max_val = np.min(src), np.max(src)
    if max_val - min_val < epsilon:
        max_val += epsilon

    return (src - min_val) / (max_val - min_val)


def normalize_rgb(img, mean, std):
    """ Normalize an image.

    Subtract mean per channel and divide by std per channel. (support
    grayscale image and RGB image).

    Args:
        img (ndarray): image to be normalized.
        mean (tuple[float] or float): mean values.
        std (tuple[float] or float): standard deviations.

    Return:
        (ndarray): the normalized RGB image.

    Note:
        For grayscale image, first rescale intensity to [0.0, 255.0].
        Then convert into a 3-channel RGB image before normalization.
    """
    img = img.astype(np.float32)

    if img.ndim == 2:
        img = normalize_grayscale(img, to_float=False) * 255.0
        img = mv.gray2rgb(img)

    return (img - mean) / std



import torch
from .average_meter import AverageMeter
from .hook import Hook
import medvision as mv


class Runner:
    def __init__(self,
                 mode,
                 model,
                 batch_processor,
                 train_dataloader=None,
                 val_dataloader=None,
                 optimizer=None,
                 work_dir=None,
                 max_epochs=10000):
        """ A training helper for PyTorch.

        Args:
            model (`torch.nn.Module`): The model to be run.
            mode ('ModeKey'): running mode.
            batch_processor (callable): A callable method that process a data
                batch. The interface of this method should be
                `batch_processor(model, data, train_mode) -> dict`
            train_dataloader ('DataLoader'): train data loader.
            val_dataloader ('DataLoader'): validation data loader.
            optimizer (dict or `Optimizer`): If it is a dict, runner will
                construct an optimizer according to it.
            work_dir (str, optional): The working directory to save
                checkpoints, logs and other outputs.
            max_epochs (int): Total training epochs.
        """
        assert isinstance(mode, mv.ModeKey)
        assert isinstance(model, torch.nn.Module)
        assert callable(batch_processor)
        assert isinstance(optimizer, (str, torch.optim.Optimizer))
        assert isinstance(work_dir, str) or work_dir is None
        assert isinstance(max_epochs, int)

        self.mode = mode
        self.epoch_runner = getattr(self, mode.value)
        self.model = model
        self.batch_processor = batch_processor
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.optimizer = self.build_optimizer(optimizer)

        # create work_dir
        self.work_dir = mv.abspath(work_dir if work_dir is not None else '.')
        mv.mkdirs(self.work_dir)

        # init TensorboardX visualizer and dataloader
        if mode == mv.ModeKey.TRAIN:
            experiment = mv.basename(self.work_dir)
            self.visualizer = mv.TensorboardVisualizer(experiment)
            self.dataloader = self.train_dataloader
        else:
            self.visualizer = None
            self.dataloader = self.val_dataloader

        # init hooks and average meter
        self._hooks = []
        self.average_meter = AverageMeter()

        # init loop parameters
        self._epoch = 0
        self._max_epochs = max_epochs if mode == mv.ModeKey.TRAIN else 1
        self._inner_iter = 0
        self._iter = 0
        self._max_iters = 0

        # get model name from model class
        if hasattr(self.model, 'module'):
            self._model_name = self.model.module.__class__.__name__
        else:
            self._model_name = self.model.__class__.__name__

    def register_hook(self, hook, priority='NORMAL'):
        """Register a hook into the hook list.

        Args:
            hook (:obj:`Hook`): The hook to be registered.
            priority (int or str or :obj:`Priority`): Hook priority.
                Lower value means higher priority.
        """
        assert isinstance(hook, Hook)

        hook.priority = mv.get_priority(priority)
        # insert the hook to a sorted list
        for i in range(len(self._hooks) - 1, -1, -1):
            if priority >= self._hooks[i].priority:
                self._hooks.insert(i + 1, hook)
                break
        else:
            self._hooks.insert(0, hook)

    def call_hook(self, fn_name):
        for hook in self._hooks:
            getattr(hook, fn_name)(self)

    def train(self, **kwargs):
        self.model.train()
        self._max_iters = self._max_epochs * len(self.train_dataloader)
        self.call_hook('before_train_epoch')

        for i, data_batch in enumerate(self.train_dataloader):
            self._inner_iter = i
            self.call_hook('before_train_iter')
            outputs = self.batch_processor(
                self.model, data_batch, is_train=True, **kwargs)
            if not isinstance(outputs, dict):
                raise TypeError('batch_processor() must return a dict')
            if 'log_vars' in outputs:
                self.average_meter.update(
                    outputs['log_vars'],
                    outputs['num_samples'])
            self.outputs = outputs
            self.call_hook('after_train_iter')
            self._iter += 1

        self.call_hook('after_train_epoch')
        self._epoch += 1

    def val(self, **kwargs):
        self.model.eval()
        self.call_hook('before_val_epoch')

        for i, data_batch in enumerate(self.val_dataloader):
            self._inner_iter = i
            self.call_hook('before_val_iter')
            with torch.no_grad():
                outputs = self.batch_processor(
                    self.model, data_batch, is_train=False, **kwargs)
            if not isinstance(outputs, dict):
                raise TypeError('batch_processor() must return a dict')
            if 'log_vars' in outputs:
                self.average_meter.update(
                    outputs['log_vars'],
                    outputs['num_samples'])
            self.outputs = outputs
            self.call_hook('after_val_iter')

        self.call_hook('after_val_epoch')
        self._epoch += 1

    def run(self, **kwargs):
        self.call_hook('before_run')
        while self.epoch < self.max_epochs:
            self.epoch_runner(self.dataloader, **kwargs)
        self.call_hook('after_run')

    def build_optimizer(self, optimizer):
        """ Init the optimizer.

        Args:
            optimizer (dict or `Optimizer`): Either an optimizer object
                or a dict used for constructing the optimizer.

        Returns:
            (`Optimizer`): An optimizer object.

        Examples:
            >>> optimizer = dict(type='SGD', lr=0.01, momentum=0.9)
            >>> type(runner.init_optimizer(optimizer))
            <class 'torch.optim.sgd.SGD'>
        """
        if isinstance(optimizer, dict):
            optimizer = mv.build_object_from_dict(
                optimizer, torch.optim, dict(params=self.model.parameters()))
        elif not isinstance(optimizer, torch.optim.Optimizer):
            raise TypeError(
                'optimizer must be either an Optimizer object or a dict, '
                'but got {}'.format(type(optimizer)))
        return optimizer

    def current_lr(self):
        """Get current learning rates.

        Returns:
            (list): Current learning rate of all param groups.
        """
        if self.optimizer is None:
            raise RuntimeError(
                'lr is not applicable because optimizer does not exist.')
        return [group['lr'] for group in self.optimizer.param_groups]

    @property
    def hooks(self):
        return self._hooks

    @property
    def model_name(self):
        return self._model_name

    @property
    def epoch(self):
        return self._epoch

    @property
    def max_epochs(self):
        return self._max_epochs

    @property
    def inner_iter(self):
        return self._inner_iter

    @property
    def iter(self):
        return self._iter

    @property
    def max_iters(self):
        return self._max_iters

    def register_logger_hooks(self):
        pass



from collections import OrderedDict
import torch
import medvision as mv


def _weights_to_cpu(state_dict_gpu):
    """ Copy a model state_dict to cpu.

    Args:
        state_dict_gpu (OrderedDict): model weights on GPU.

    Return:
        (OrderedDict): model weights on CPU.
    """
    state_dict_cpu = OrderedDict()
    for key in state_dict_gpu:
        state_dict_cpu[key] = state_dict_gpu[key].cpu()
    return state_dict_cpu


def save_checkpoint(model, path, optimizer=None, metadata=None):
    """ Save checkpoint to file.

    The checkpoint will have 3 fields: ``metadata``, ``state_dict`` and
    ``optimizer``.

    Args:
        model (Module): module whose params are to be saved.
        path (str): path to save the checkpoint file.
        optimizer ('Optimizer', optional): optimizer to be saved.
        metadata (dict, optional): metadata to be saved in checkpoint.
    """
    assert isinstance(metadata, (dict, type(None)))
    if metadata is None:
        metadata = {}

    mv.mkdirs(mv.parentdir(path))

    # if wrapped by nn.DataParallel, remove the wrapper
    if hasattr(model, 'module'):
        model = model.module

    # make a checkpoint
    checkpoint = {'state_dict': _weights_to_cpu(model.state_dict())}
    if optimizer is not None:
        checkpoint['optimizer'] = optimizer.state_dict()
    if metadata is not None:
        checkpoint['metadata'] = metadata

    torch.save(checkpoint, path)


def load_checkpoint(model, path, map_location='cpu', strict=True):
    """ Load checkpoint from a file.

    Args:
        model (Module): module to load checkpoint.
        path (str): checkpoint weights file path.
        map_location (str): same as :func:`torch.load`.
        strict (bool): whether to allow different params for the model and
            checkpoint.

    TODO: support load checkpoint from url with torch.utils.model_zoo
    """
    # load checkpoint from file
    checkpoint = torch.load(path, map_location=map_location)

    # get state_dict from checkpoint
    if isinstance(checkpoint, OrderedDict):
        # checkpoint saved by torch.save(model.state_dict(), path)
        state_dict = checkpoint
    elif isinstance(checkpoint, dict) and 'state_dict' in checkpoint:
        # checkpoint saved by save_checkpoint()
        state_dict = checkpoint['state_dict']
    else:
        raise RuntimeError(
            'No state_dict found in checkpoint file {}'.format(path))

    # strip prefix of state_dict
    if list(state_dict.keys())[0].startswith('module.'):
        state_dict = {k[7:]: v for k, v in state_dict.items()}

    # load state_dict
    if hasattr(model, 'module'):
        model.module.load_state_dict(state_dict, strict)
    else:
        model.load_state_dict(state_dict, strict)


def save_ckpt_to_dir(model, ckpt_dir, identifier):
    """ Save checkpoint to file and make a soft link to the latest ckpt.

    Args:
        model (Module): module whose params are to be saved.
        ckpt_dir (str): directory to save the checkpoint file.
        identifier (str): ckpt identifier (e.g. number of epochs).
    """
    model_path = mv.joinpath(ckpt_dir, str(identifier) + '.pth')
    mv.save_checkpoint(model, model_path)

    link_path = mv.joinpath(ckpt_dir, 'latest' + '.pth')
    mv.symlink(mv.abspath(model_path), link_path)


def load_ckpt_from_dir(model, ckpt_dir, identifier='latest'):
    """ Load checkpoint from a directory with given identifier.

    Args:
        model (Module): module whose params are to be saved.
        ckpt_dir (str): directory to load the checkpoint file.
        identifier (str): ckpt identifier (e.g. number of epochs).
    """
    model_path = mv.joinpath(ckpt_dir, str(identifier) + '.pth')
    mv.load_checkpoint(model, model_path)



from collections import OrderedDict
import numpy as np
from .compute_overlap import compute_overlap


def _compute_ap_voc07(rec, pre):
    ap = 0.
    for t in np.arange(0., 1.1, 0.1):
        if np.sum(rec >= t) == 0:
            p = 0
        else:
            p = np.max(pre[rec >= t])
        ap = ap + p / 11.

    return ap


def _compute_ap_voc12(rec, pre):
    # correct AP calculation
    # first append sentinel values at the end
    mrec = np.concatenate(([0.], rec, [1.]))
    mpre = np.concatenate(([0.], pre, [0.]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # to calculate area under PR curve, look for points
    # where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * precision
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])

    return ap


def _compute_ap(rec, pre, use_voc07_metric=False):
    """ Compute the average precision, given the recall and precision curves.

    Code originally from https://github.com/rbgirshick/py-faster-rcnn.

    Args:
        rec (list): the recall curve .
        pre (list): the precision curve.
        use_voc07_metric (bool): whether to voc2007 11 points metric.

    Returns:
        The average precision as computed in py-faster-rcnn.
    """
    if use_voc07_metric:
        return _compute_ap_voc07(rec, pre)
    else:
        return _compute_ap_voc12(rec, pre)


def _to_list(dts, gts):
    """ Convert gt/dt from OrderedDict to list[list[ndarray]].
    Args:
        dts (OrderedDict or list[list[ndarray]]): detected bounding boxes for
            different labels in a set of images, each bbox is of shape (n, 5).
        gts (OrderedDict or list[list[ndarray]]): ground truth bounding boxes
            for different labels in a set of images, each bbox is of
            shape (n, 4).
            gts[img_id][label_id] = bboxes (for a specific label in an image).
    """
    if isinstance(dts, OrderedDict):
        dts = [value for key, value in dts.items()]

    if isinstance(gts, OrderedDict):
        gts = [value for key, value in gts.items()]

    return dts, gts


def eval_det4binarycls(dts, gts, score_thr=0.05):
    """ Evaluate a detector's classification capability.

    Only support 1-class detection. If (at least) 1 target is detected in
    an image, the image is considered to be 'positive'. Otherwise, the image
    is considered to be 'negative'.

    Args:
        dts (OrderedDict or list[list[ndarray]]): detected bounding boxes for
            different labels in a set of images, each bbox is of shape (n, 5).
        gts (OrderedDict or list[list[ndarray]]): ground truth bounding boxes
            for different labels in a set of images, each bbox is of
            shape (n, 4).
            gts[img_id][label_id] = bboxes (for a specific label in an image).
        score_thr (float): score confidence threshold to determine whether a
            detection is valid.
    Returns:
        (dict): a dict containing classification metrics TP, FP, TN, FN,
        accuracy, recall, precision.
    """
    assert len(gts) == len(dts)
    dts, gts = _to_list(dts, gts)

    # compute detector's classification capability
    tp, fp, tn, fn = 0, 0, 0, 0
    for i in range(len(gts)):
        assert len(gts[i]) == 1, 'only support 1-class detection'
        assert len(dts[i]) == 1, 'only support 1-class detection'
        gt = gts[i][0]
        dt = dts[i][0]

        if len(dt) != 0:
            has_dt = dt[:, 4].max() > score_thr
        else:
            has_dt = False

        has_gt = (len(gt) != 0)

        if has_dt and has_gt:
            tp += 1
        elif has_dt and not has_gt:
            fp += 1
        elif not has_dt and has_gt:
            fn += 1
        else:  # not has_dt and not has_gt:
            tn += 1

    # build result
    result = OrderedDict()

    result['tp'] = tp
    result['fp'] = fp
    result['tn'] = tn
    result['fn'] = fn

    eps = np.finfo(np.float32).eps
    result['accuracy'] = (tp + tn) / np.maximum(tp + fn + tn + fp, eps)
    result['sensitivity'] = tp / np.maximum(tp + fn, eps)
    result['specificity'] = tn / np.maximum(tn + fp, eps)
    result['recall'] = result['sensitivity']
    result['precision'] = tp / np.maximum(tp + fp, eps)

    return result


def eval_det(dts, gts, num_classes=1, iou_thr=0.5, score_thr=0.05):
    """ Evaluate a given dataset by comparing DT with GT.

    Args:
        dts (OrderedDict or list[list[ndarray]]): detected bounding boxes for
            different labels in a set of images, each bbox is of shape (n, 5).
        gts (OrderedDict or list[list[ndarray]]): ground truth bounding boxes
            for different labels in a set of images, each bbox is of
            shape (n, 4).
            gts[img_id][label_id] = bboxes (for a specific label in an image).
        num_classes (int): number of classes to detect.
        iou_thr (float): threshold to determine whether a detection is
            positive or negative.
        score_thr (float): score confidence threshold to determine whether a
            detection is valid.

    Returns:
        (OrderedDict): AP, number of GT bboxes, FROC curve for each label.
    """
    assert len(gts) == len(dts)
    dts, gts = _to_list(dts, gts)
    num_imgs = len(gts)

    results = OrderedDict()

    # match ground truths and detection results
    for label in range(num_classes):
        fps = np.zeros((0,), np.float32)
        tps = np.zeros((0,), np.float32)
        scores = np.zeros((0,), np.float32)
        num_anns = 0.0

        for i in range(num_imgs):
            gt = gts[i][label]
            dt = dts[i][label]
            num_anns += len(gt)
            matched_anns = []

            for d in dt:
                scores = np.append(scores, d[4])

                if len(gt) == 0:
                    fps = np.append(fps, 1)
                    tps = np.append(tps, 0)
                    continue

                overlaps = compute_overlap(np.expand_dims(d, axis=0), gt)
                matched_ann = np.argmax(overlaps, axis=1)
                max_overlap = overlaps[0, matched_ann]

                if (max_overlap > iou_thr and
                        matched_ann not in matched_anns):
                    fps = np.append(fps, 0)
                    tps = np.append(tps, 1)
                    matched_anns.append(matched_ann)
                else:
                    fps = np.append(fps, 1)
                    tps = np.append(tps, 0)

        # no annotations -> AP for this class is None
        if num_anns == 0:
            results[label] = OrderedDict()
            results[label]['ap'] = None
            results[label]['num_gt_bboxes'] = 0
            results[label]['froc'] = None
            continue

        # sort by score
        indices = np.argsort(-scores)
        fps, tps = fps[indices], tps[indices]

        # compute false positives and true positives at retrieval cutoff
        # of k bboxes (k in [1, n], n is the total number of detected bboxes)
        fps, tps = np.cumsum(fps), np.cumsum(tps)

        # compute recall and precision
        recall = tps / num_anns
        precision = tps / np.maximum(tps + fps, np.finfo(np.float32).eps)

        # compute AP, number of ground truth bboxes and FROC
        results[label] = OrderedDict()
        results[label]['ap'] = _compute_ap(recall, precision)
        results[label]['num_gt_bboxes'] = num_anns
        results[label]['froc'] = fps / num_imgs, recall  # recall = sensitivity

    return results



from .hook import Hook


class LoggerHook(Hook):
    """ Base class for logger hooks.

    Args:
        interval (int): Logging interval (every k iterations).
    """
    def __init__(self, interval=10):
        self.interval = interval
        self.visualizer = None

    def log(self, runner):
        pass

    def before_epoch(self, runner):
        runner.average_meter.clear()  # clear logs of last epoch

    def after_train_iter(self, runner):
        if self.every_n_inner_iters(runner, self.interval):
            runner.log_buffer.average(self.interval)

        if runner.average_meter.ready:
            self.log(runner)

    def after_train_epoch(self, runner):
        if runner.average_meter.ready:
            self.log(runner)

    def after_val_epoch(self, runner):
        runner.average_meter.average()
        self.log(runner)



import matplotlib.pyplot as plt
import sklearn


def draw_froc(average_fps, sensitivity, save_path=None, **kwargs):
    """ Plot the FROC curve.

    Args:
        average_fps (list): A list containing the average number of false
        positives per image for different thresholds.
        sensitivity (list):  A list containing overall sensitivity for
        different thresholds.
        save_path (str): path to save the froc curve drawing.
    """
    plt.figure()
    plt.xlabel('False Positives')
    plt.ylabel('Sensitivity')
    plt.title('FROC Curve')
    plt.ylim([0, 1])
    plt.plot(average_fps, sensitivity, **kwargs)
    plt.grid()
    if save_path is not None:
        plt.savefig(save_path)
    plt.show()


def save_roc_curve(roc_curve, save_path=None):
    fpr, tpr, _ = roc_curve
    roc_auc = sklearn.metrics.auc(fpr, tpr)
    plt.figure()
    plt.plot(fpr, tpr, color='darkorange', lw=2,
             label='ROC curve (auc = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.xlabel('FPR')
    plt.ylabel('TPR')
    plt.xticks()
    plt.yticks()
    plt.title('ROC Curve')
    plt.legend(loc="lower right")
    if save_path is not None:
        plt.savefig(save_path)
    plt.show()


def save_pr_curve(pr_curve, save_path=None):
    precision, recall, _ = pr_curve
    plt.plot(recall, precision, color='darkorange', lw=2)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.xticks()
    plt.yticks()
    plt.title('PR Curve')
    if save_path is not None:
        plt.savefig(save_path)
    plt.show()



from torch.utils.data import DataLoader
from .cls_dataset import ClsDataset
import medvision as mv


def build_cls_dataloader(cfg, mode, build_transform, image_loader):
    # TODO: avoid using cfg as argument
    # init dataloader parameters
    if mode == mv.ModeKey.TRAIN:
        batch_size = cfg.TRAIN.BATCH_SIZE
        shuffle = True
        num_workers = cfg.TRAIN.NUM_WORKERS
        pin_memory = True
    else:
        batch_size = cfg.TEST.BATCH_SIZE
        shuffle = False
        num_workers = cfg.TEST.NUM_WORKERS
        pin_memory = False

    # build dataset
    cls_dataset = ClsDataset(
        cfg=cfg,
        mode=mode,
        build_transform=build_transform,
        image_loader=image_loader
    )

    # build dataloader
    data_loader = DataLoader(
        cls_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=num_workers,
        drop_last=False,
        pin_memory=pin_memory
    )

    return data_loader



from enum import Enum


class Priority(Enum):
    """ priority levels.
    +------------+------------+
    | Level      | Value      |
    +============+============+
    | HIGHEST    | 0          |
    +------------+------------+
    | VERY_HIGH  | 10         |
    +------------+------------+
    | HIGH       | 30         |
    +------------+------------+
    | NORMAL     | 50         |
    +------------+------------+
    | LOW        | 70         |
    +------------+------------+
    | VERY_LOW   | 90         |
    +------------+------------+
    | LOWEST     | 100        |
    +------------+------------+
    """
    HIGHEST = 0
    VERY_HIGH = 10
    HIGH = 30
    NORMAL = 50
    LOW = 70
    VERY_LOW = 90
    LOWEST = 100


def get_priority(priority):
    """ Get priority value.

    Args:
        priority (int or str or :obj:`Priority`): Priority.

    Returns:
        (int): The priority value.
    """
    if isinstance(priority, int):
        if priority < 0 or priority > 100:
            raise ValueError('priority must be between 0 and 100')
        return priority
    elif isinstance(priority, Priority):
        return priority.value
    elif isinstance(priority, str):
        return Priority[priority.upper()].value
    else:
        raise TypeError('unsupported priority type')



import sys


def build_object_from_dict(dictionary, parent=None, default_args=None):
    """ Initialize an object from dict.

    The dict must contain the key "type", which indicates the object type, it
    can be either a string or type, such as "list" or ``list``. Remaining
    fields are treated as the arguments for constructing the object.

    Args:
        dictionary (dict): Object types and arguments.
        parent (:class:`module`): Module which may containing expected object
            classes.
        default_args (dict, optional): Default arguments for initializing the
            object.

    Returns:
        (any type): Object built from the dict.
    """
    assert isinstance(dictionary, dict) and 'type' in dictionary
    assert isinstance(default_args, dict) or default_args is None

    args = dictionary.copy()
    object_type = args.pop('type')

    if isinstance(object_type, str):
        if parent is not None:
            object_type = getattr(parent, object_type)
        else:
            object_type = sys.modules[object_type]
    elif not isinstance(object_type, type):
        raise TypeError('unsupported type {}'.format(type(object_type)))

    if default_args is not None:
        for name, value in default_args.items():
            args.setdefault(name, value)

    return object_type(**args)



import collections
from enum import Enum, unique
import math
import cv2
import matplotlib.pyplot as plt
import numpy as np
import medvision as mv


@unique
class Color(Enum):
    """ An enum that defines common colors.
    """
    red = (0, 0, 255)
    green = (0, 255, 0)
    blue = (255, 0, 0)
    cyan = (255, 255, 0)
    yellow = (0, 255, 255)
    magenta = (255, 0, 255)
    white = (255, 255, 255)
    black = (0, 0, 0)


def _imshow_tight(img, title):
    cmap = 'gray' if img.ndim == 2 else None
    plt.imshow(img, cmap=cmap)
    plt.axis('off')
    plt.xlim([0, img.shape[1]])
    plt.ylim([img.shape[0], 0])
    plt.title(title)


class Switcher:
    def __init__(self, ax, imgs):
        self.ax = ax

        self.imgs = imgs
        self.num_imgs = len(imgs)
        self.index = 0

        self.im = ax.imshow(imgs[self.index])
        self.update()

    def on_press(self, event):
        self.index = (self.index + 1) % self.num_imgs
        self.update()

    def update(self):
        self.im.set_data(self.imgs[self.index])
        self.im.axes.figure.canvas.draw_idle()
        try:
            self.im.axes.figure.canvas.flush_events()
        except NotImplementedError:
            pass


def _imshow_switcher(imgs, title=''):
    """ imshow with channel changer.

    An enhanced version of 'imshow'. Images (stored in 'imgs') can be
    switched by single clicking.

    Args:
        imgs (list[ndarray]): images to be displayed.
        title (str): title of the plot.
    """
    fig, ax = plt.subplots(num=title)
    fig.tight_layout()
    ax.set_title(title)
    ax.axis('off')
    ref_img = imgs[0]
    ax.set_xlim([0, ref_img.shape[1]])
    ax.set_ylim([ref_img.shape[0], 0])
    switcher = Switcher(ax, imgs)
    fig.canvas.mpl_connect("button_press_event", switcher.on_press)
    plt.show()


def imshow(imgs, num_cols=None, fig_name=None, titles='', save_path=None):
    """ Show an image or multiple images in a single canvas.

    Args:
        imgs (ndarray or tuple/list[ndarray]): images to be shown.
        num_cols (int): image number per column for multiple images display.
            If not given, this parameter is automatically determined.
        fig_name (str): name of the plot.
        titles (str or list[str]): sub-plot titles.
        save_path (str, optional): path to save the image.
    """
    if not isinstance(imgs, collections.Sequence):
        imgs = [imgs]
    num_imgs = len(imgs)

    if isinstance(titles, str):
        titles = [titles] * num_imgs
    assert len(titles) == num_imgs

    if num_cols is None:
        num_cols = int(math.ceil(math.sqrt(num_imgs)))
    assert num_cols <= len(imgs)
    num_rows = int((num_imgs + num_cols - 1) // num_cols)

    plt.figure(fig_name)
    plt.tight_layout()

    for i, img in enumerate(imgs):
        plt.subplot(num_rows, num_cols, i + 1)
        _imshow_tight(img, titles[i])

    # mng = plt.get_current_fig_manager()
    # mng.resize(*mng.window.maxsize())

    if save_path is not None:
        plt.savefig(save_path)

    plt.show()


def imshow_bboxes(img,
                  bboxes,
                  score_thr=0,
                  colors=Color.green,
                  top_k=-1,
                  thickness=1,
                  font_scale=0.5,
                  show=True,
                  title='',
                  save_path=None):
    """ Draw bounding boxes on an image.

    To display detection result or compare detection results by different
    algorithms.

    Args:
        img (str or ndarray): image (or file path) to be displayed.
        bboxes (list or ndarray): a list of ndarray of shape (k, 4) or (n, 5).
        score_thr (float): minimum score of bboxes to be shown.
        colors (Color or list[Color]): color or list of colors.
        top_k (int): plot the first k bboxes only if set positive.
            Otherwise, plot all the bboxes.
        thickness (int): line thickness.
        font_scale (float): font scales of texts.
        show (bool): True: show the image; False: save the image.
        title (str): title of the plot.
        save_path (str, optional): path to save the image.
    """
    if isinstance(img, str):
        img = mv.imread(img)
    else:
        img = img if img.ndim == 3 else mv.gray2rgb(img)

    if isinstance(bboxes, np.ndarray):
        bboxes = [bboxes]

    if isinstance(colors, Color):
        colors = [colors] * len(bboxes)

    assert len(bboxes) == len(colors)

    plot_prob = True if bboxes[0].shape[1] == 5 else False

    if score_thr > 0:
        for i in range(len(bboxes)):
            assert bboxes[i].shape[1] == 5
            scores = bboxes[i][:, -1]
            indices = scores > score_thr
            bboxes[i] = bboxes[i][indices, :]

    img_with_result = img.copy()
    for i, _bboxes in enumerate(bboxes):
        _bboxes_int = _bboxes.astype(np.int32)
        if top_k <= 0:
            _top_k = _bboxes.shape[0]
        else:
            _top_k = min(top_k, _bboxes.shape[0])
        for j in range(_top_k):
            left_top = (_bboxes_int[j, 0], _bboxes_int[j, 1])
            right_bottom = (_bboxes_int[j, 2], _bboxes_int[j, 3])
            cv2.rectangle(img_with_result, left_top, right_bottom,
                          colors[i].value, thickness)
            if plot_prob:
                label_text = '%.2f' % _bboxes[j, -1]
                cv2.putText(img_with_result, label_text,
                            (_bboxes_int[j, 0], _bboxes_int[j, 1] - 2),
                            cv2.FONT_HERSHEY_COMPLEX, font_scale,
                            colors[i].value,
                            thickness)

    if show:
        _imshow_switcher([img_with_result, img], title)

    if save_path is not None:
        mv.imwrite(save_path, img_with_result)


if __name__ == '__main__':
    im = cv2.imread('../../tests/data/pngs/Blue-Ogi.png', cv2.IMREAD_GRAYSCALE)
    imshow(im, fig_name='show single image', titles='name')

    h, w = im.shape[:2]
    bbox = np.array([w // 3, h // 3, w * 2 // 3, h * 2 // 3,
                     0.5]).reshape(-1, 5)
    imshow_bboxes(im, bbox, score_thr=0.2, title='draw bounding boxes')

    im = cv2.imread('../../tests/data/pngs/Blue-Ogi.png', cv2.IMREAD_COLOR)
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    imshow([im] * 5, fig_name='show multiple images',
           titles=[str(i) for i in range(5)])



import json
import cv2
import numpy as np


def load_bdc_dr_annot(annot_path, class2label=None):
    """ Load a DR BDC annotation file.

    Args:
        annot_path (str): annotation file path.
        class2label (dict or callable): convert classname to label. If
            it is not dict or callable, the classname will not be converted.

    Returns:
        (list[ndarray]): a list containing all annotation contours.
    """
    with open(annot_path, 'rt', encoding='GBK') as fd:
        lines = fd.readlines()

    contours = []
    for line in lines:
        classname, content = line.split('|')

        if isinstance(class2label, dict):
            label = class2label[classname]
        elif callable(class2label):
            label = class2label(classname)
        else:
            label = classname

        contour = np.array(json.loads(content)).astype(np.int)
        contour = contour[:, :2]
        contour = np.reshape(contour, (-1, 1, 2))
        contours.append((label, contour))

    return contours


def _contour2bbox(contour):
    x, y, w, h = cv2.boundingRect(contour)
    return [x, y, x + w, y + h]


def load_bdc_dr_bbox(annot_path, class2label=None):
    """ Load a 2D BDC annotation file and convert it to a bounding box.

    Args:
        annot_path (str): annotation file path.
        class2label (dict or callable): convert classname to label. If
            it is not dict or callable, the classname will not be converted.

    Returns:
        (list[ndarray]): a list containing converted annotation bboxes.
    """
    bboxes = []

    contours = load_bdc_dr_annot(annot_path, class2label)
    for label, contour in contours:
        bbox = _contour2bbox(contour)
        bboxes.append((label, bbox))

    return bboxes


if __name__ == '__main__':
    # TODO: move to unittest
    image_path = '/mnt/sdb1/tb/internal/tb/002565.dcm'
    annot_path = '/mnt/sdb1/tb/internal/label/002565.txt'
    bboxes = load_bdc_dr_bbox(annot_path, lambda x: 1)
    import medvision as mv
    image = mv.dcmread_dr(image_path)
    image = mv.normalize_grayscale(image) * 255.0
    image = image.astype(np.uint8)
    tmp = [bbox for label, bbox in bboxes]
    tmp = np.array(tmp)
    mv.imshow_bboxes(image, tmp, thickness=4)
    print(bboxes)



# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Sergey Karayev
# --------------------------------------------------------

import numpy as np
cimport numpy as np


def compute_overlap(
    np.ndarray[float, ndim=2] boxes,
    np.ndarray[float, ndim=2] query_boxes
):
    """
    Args
        a: (N, 4) or (N, 5) ndarray of float
        b: (K, 4) or (N, 5) ndarray of float

    Returns
        overlaps: (N, K) ndarray of overlap between boxes and query_boxes
    """
    cdef unsigned int N = boxes.shape[0]
    cdef unsigned int K = query_boxes.shape[0]
    cdef np.ndarray[float, ndim=2] overlaps = np.zeros((N, K), dtype=np.float32)
    cdef float iw, ih, box_area
    cdef float ua
    cdef unsigned int k, n
    for k in range(K):
        box_area = (
            (query_boxes[k, 2] - query_boxes[k, 0] + 1) *
            (query_boxes[k, 3] - query_boxes[k, 1] + 1)
        )
        for n in range(N):
            iw = (
                min(boxes[n, 2], query_boxes[k, 2]) -
                max(boxes[n, 0], query_boxes[k, 0]) + 1
            )
            if iw > 0:
                ih = (
                    min(boxes[n, 3], query_boxes[k, 3]) -
                    max(boxes[n, 1], query_boxes[k, 1]) + 1
                )
                if ih > 0:
                    ua = float(
                        (boxes[n, 2] - boxes[n, 0] + 1) *
                        (boxes[n, 3] - boxes[n, 1] + 1) +
                        box_area - iw * ih
                    )
                    overlaps[n, k] = iw * ih / ua
    return overlaps



from enum import Enum, unique
import numpy as np
import torch


@unique
class ModeKey(Enum):
    TRAIN = 'train'
    VAL = 'val'
    TEST = 'test'


def nograd(f):
    def decorator(*args, **kwargs):
        with torch.no_grad():
            return f(*args, **kwargs)
    return decorator


def make_np(x):
    if isinstance(x, np.ndarray):
        return x

    if 'torch' in str(type(x)):
        return x.cpu().numpy()

    raise NotImplementedError(
        'Got {}, but expected numpy array or torch tensor.'.format(type(x)))



from .checkpoint import (save_checkpoint, load_checkpoint,
                         save_ckpt_to_dir, load_ckpt_from_dir)
from .dataset.cls_dataloader import build_cls_dataloader
from .runner import Hook, LoggerHook, ClsLoggerHook, Runner
from .util import ModeKey, nograd, make_np


__all__ = [
    'save_checkpoint', 'load_checkpoint',
    'save_ckpt_to_dir', 'load_ckpt_from_dir',
    'build_cls_dataloader',
    'Hook', 'LoggerHook', 'ClsLoggerHook', 'Runner',
    'ModeKey', 'nograd', 'make_np'
]



from .transforms import (rgb2gray, gray2rgb,
                         vflip, hflip, rot90, rotate, resize, rescale, crop,
                         center_crop, pad_to_square,
                         normalize_grayscale, normalize_rgb)

from .io import ImreadMode, imread, imwrite

__all__ = [
    'rgb2gray', 'gray2rgb',
    'vflip', 'hflip', 'rot90', 'rotate', 'resize', 'rescale', 'crop',
    'center_crop', 'pad_to_square',
    'normalize_grayscale', 'normalize_rgb',
    'ImreadMode', 'imread', 'imwrite'
]



from .dsmd import load_dsmd, save_dsmd, load_c2l
from .classification import (gen_cls_dsmd_file_from_datafolder,
                             gen_cls_ds_from_datafolder)
from .util import make_dsmd, split_dsmd_file, match_dsmds


__all__ = [
    'load_dsmd', 'save_dsmd', 'load_c2l',

    'gen_cls_dsmd_file_from_datafolder', 'gen_cls_ds_from_datafolder',

    'make_dsmd', 'split_dsmd_file', 'match_dsmds'
]



import time
import numpy as np
from tensorboardX import SummaryWriter
import medvision as mv


class TensorboardVisualizer:
    """ A wrapper class for tensorboardX.

    Note:
        Original tensorboardX API call is supported.
    """
    def __init__(self, experiment=None, **kwargs):
        comment = '_' + str(experiment) if experiment is not None else ''
        if self.experiment_exists(experiment):
            self.has_writer = False
            raise ValueError(
                'experiment [{}] already exists.'.format(experiment))
        else:
            self.has_writer = True
        self.writer = SummaryWriter(comment=comment, **kwargs)

    def __del__(self):
        if self.has_writer:
            self.writer.close()

    @staticmethod
    def experiment_exists(experiment):
        if not mv.isdir('runs'):
            return False

        all_experiments = mv.listdir('runs')
        all_experiments = [e.split('_', 3)[-1] for e in all_experiments]
        if experiment in all_experiments:
            return True
        else:
            return False

    def plot(self, name, x, y, group='data'):
        tag = group + '/' + name
        self.writer.add_scalar(tag, y, x)

    def imshow(self, name, img_tensor, global_step=None, group='image'):
        """ Accept NCHW numpy.ndarray or torch tensor as input.
        """
        imgs = mv.make_np(img_tensor)
        assert imgs.ndim == 4

        if imgs.shape[1] == 1:
            imgs = np.concatenate([imgs, imgs, imgs], 1)

        tag = group + '/' + name
        self.writer.add_images(tag, imgs, global_step)

    def log(self, info, global_step=None, tag='text/log'):
        text_string = '[{time}] {info} <br>'.format(
            time=time.strftime('%y-%m-%d %H:%M:%S'),
            info=info
        )
        self.writer.add_text(tag, text_string, global_step)

    def __getattr__(self, name):
        return getattr(self.writer, name)


if __name__ == '__main__':
    import torch
    from torchvision import datasets

    visualizer = TensorboardVisualizer('tensorboard-test')

    for n_iter in range(100):
        dummy_s1 = torch.rand(1)
        dummy_s2 = torch.rand(1)
        # data grouping by `slash`
        visualizer.plot('scalar1', n_iter, dummy_s1[0])
        visualizer.plot('scalar2', n_iter, dummy_s2[0])

        dummy_img = torch.rand(32, 1, 64, 64)  # output from network
        dummy_img_c = torch.rand(32, 3, 64, 64)  # output from network
        if n_iter % 10 == 0:
            visualizer.imshow('feature_map', dummy_img, n_iter)
            visualizer.imshow('feature_map2', dummy_img_c, n_iter)

            visualizer.log(str(n_iter), n_iter)

            # needs tensorboard 0.4RC or later
            visualizer.add_pr_curve(
                'xoxo', np.random.randint(2, size=100),
                np.random.rand(100), n_iter)

    dataset = datasets.MNIST('mnist', train=False, download=True)
    images = dataset.data[:100].float()
    label = dataset.targets[:100]

    features = images.view(100, 784)
    visualizer.add_embedding(features, metadata=label,
                             label_img=images.unsqueeze(1))

    try:
        visualizer_2 = TensorboardVisualizer('tensorboard-test')
    except ValueError:
        print('duplicate experiments successfully detected')
    else:
        raise ValueError('duplicate experiments undetected')
